###
## Functions
###


## calculate the probability of the observed number of mutations in one genome
pr_one_genome <- function(test.parameter, single.genome.data, this.mutation.list) {

	mutation.sum <- 0;
	syn <- 0; 
	for (i in 1:length(this.mutation.list))
	{
		##mutation.sum = number of mutations with this signature
		mutation.sum <- mutation.sum + single.genome.data[this.mutation.list[i]][1,1]
		
		## syn = number of synonymous substitutions possible with this mutational signature
		## that would have been recovered by our data in THIS genome
		syn <- syn + single.genome.data[paste(this.mutation.list[i], ".syn", sep="")][1,1]
	}
	
  ## test.parameter represents the rate of this mutation per the number of this synonymous substitutions
  ## with this signature that are possible. It is NOT a per nucleotide rate
  
	mutation.lambda = test.parameter * single.genome.data$generation * syn;
	pr = dpois(mutation.sum, mutation.lambda)

	##need to catch decimal underflow errors
	if (pr == 0) {
		concat("Underflow error!! Adjust bounds of optimization\n")
		exit()
	}

	return(pr)
}

## calculate the probability of the observed number of mutations in all genomes
## given a single ancestral rate, returns neg log so that we can minimize this function
## (which means maximizing the likelihood)
neg_log_pr_all_genomes <- function(test.parameter, structured.data, this.mutation.list) {

	log_pr_all <- 0;
		
	for (i in 1:length(structured.data)) {
		num.genomes.in.line <- dim(structured.data[[i]])[1]
		log_pr_this_population <- 0;
		for (j in 1:num.genomes.in.line) {
			genome_pr = pr_one_genome(test.parameter, structured.data[[i]][j,], this.mutation.list)
			log_pr_this_population = log_pr_this_population + log(genome_pr);
		}
		
		## divide by the number of genomes from this line, because if we have two genomes,
		## in resampled data sets there would be a 1/2 chance of observing either, this gives geometric mean...
		log_pr_all <- log_pr_all + log_pr_this_population / num.genomes.in.line
		
	}
	return ( -log_pr_all )
}


## calculate the probability of the observed number of mutations in all genomes
## given a single ancestral rate, returns neg log so that we can minimize this function
## (which means maximizing the likelihood)

fit_ml_model <- function(structured.data, mutation.list) {
	fit.rates = rep(0, length(mutation.list));
	
	for (k in 1:length(mutation.list)) {
		mutation.sum <- 0;
		this.mutation.list = mutation.list[[k]];

		for (j in 1:length(this.mutation.list))
		{
			for (i in 1:length(structured.data)) {
				num.genomes.in.line <- dim(structured.data[[i]])[1]
				for (m in 1:num.genomes.in.line) {
				
					single.genome.data <- structured.data[[i]][m,]					
					mutation.sum <- mutation.sum + single.genome.data[this.mutation.list[j]][1,1]
				}
			}
		}
				
		if (mutation.sum == 0) {
			fit.rates[k] = 0
		} else {
			opt <- optimize(neg_log_pr_all_genomes, lower=1E-18, upper=1E-8, tol=1E-20, structured.data=structured.data, this.mutation.list=this.mutation.list)
			fit.rates[k] = opt$minimum
		}
	}
	
	#no mutations observed, can't fit
	
	return(fit.rates)
}

##
## Resample data by population WITH replacement for bootstrap
##
resample.data <- function(structured.data) {

	new.data <- vector("list", length(structured.data))
	for (i in 1:length(new.data))
	{
		new.data[[i]] = structured.data[[sample(1:length(new.data), 1)]]
	}
	
	return(new.data)
}

##
## Create a table with rates calculated for the requested number of
## bootstrap replicates.
##
bootstrap_95_CI <- function(structured.data, mutation.list, n) {

	val = c();

	for (i in 1:n)
	{
		cat("resampled dataset #", i, "\n")
		new.data = resample.data(structured.data)
		val = rbind(val, fit_ml_model(new.data, mutation.list) )
	}	
	
	return(val)
}


##
## Perform Tukey's N-1 jacknife
##
jackknife.data <- function(structured.data, tot.mutations.signature) {

	for (i in 1:length(structured.data))
	{
		new.data[[i]] = structured.data[[sample(1:length(new.data), 1)]]
		
		val = rbind(val, fit_ml_model(jackknife.data, mutation.list) )

	}
	
	

}

##
## Perform Tukey's Jacknife
##
jackknife_95_CI <- function(structured.data, mutation.list, tot.mutations.signature, outfile) {

	ml.best = c();
	ml.tot.rates = c();
	pseudovalues = c();
	n = length(structured.data);
	#cat("n = ", n, "\n");

	actual.ml.best = fit_ml_model(structured.data , mutation.list)
	actual.ml.tot.rate = 0
	for (j in 1:length(mutation.list)) {
	  	actual.ml.tot.rate <- actual.ml.tot.rate + actual.ml.best[j] * tot.mutations.signature[j];
	}
	#cat("actual.ml.tot.rate  = ", actual.ml.tot.rate , "\n"); 
	
	for (i in 1:length(structured.data))
	{
		new.data = structured.data;
		new.data[[i]] <- NULL;	
		#print(new.data)
			
		ml.best = fit_ml_model(new.data , mutation.list)
		#print(ml.best)
		
		ml.tot.rate <- 0
		for (j in 1:length(mutation.list)) {
		  	ml.tot.rate <- ml.tot.rate + ml.best[j] * tot.mutations.signature[j];
		}
		#cat( "Jacknife overall rate #", i, " = ", ml.tot.rate, "\n");
		ml.tot.rates  = rbind(ml.tot.rates, ml.tot.rate)
		
		
		pseudovalues[i] = n * actual.ml.tot.rate - (n-1) * ml.tot.rate;
	}
	#cat(pseudovalues, "\n")
	
	jackknife.estimate = mean(pseudovalues);
	
	# this sd is denominator = n
	sd(pseudovalues);
	jackknife.stderr = sd(pseudovalues) / sqrt(n-1);
	
	my.df = n - 1;
	
	cat("**** jackknife ****\n", file=outfile)
	cat("jacknife estimate = ", jackknife.estimate, "\n", file=outfile)
	cat("jacknife stderr   = ", jackknife.stderr, "\n", file=outfile)
	cat("jacknife df       = ", my.df, "\n", file=outfile)

	plusCI <- jackknife.estimate + qt(0.975, df = my.df) * jackknife.stderr
	minusCI <- jackknife.estimate + qt(0.025, df = my.df) * jackknife.stderr

	cat("95% confidence limits [", minusCI, ",", plusCI, "]", "\n", file=outfile)
}


###
## Main code
###

input.name = "mutation_rate_data.tab"
output.name =  "ancestral_mutation_rate_maximum_likelihood_fit.tab"


#### Get arguments!

for (e in commandArgs()) {
  ta = strsplit(e,"=",fixed=TRUE)
  if(! is.na(ta[[1]][2])) {
    temp = ta[[1]][2]
    if(substr(ta[[1]][1],nchar(ta[[1]][1]),nchar(ta[[1]][1])) == "I") {
      temp = as.integer(temp)
    }
    if(substr(ta[[1]][1],nchar(ta[[1]][1]),nchar(ta[[1]][1])) == "N") {
      temp = as.numeric(temp)
    }
    assign(ta[[1]][1],temp)
    cat("assigned ",ta[[1]][1]," the value of |",temp,"|\n")
  } else {
    assign(ta[[1]][1],TRUE)
    cat("assigned ",ta[[1]][1]," the value of TRUE\n")
  }
}

#####

num.bootstrap.resamplings = 3

##transition/transversion division
transition.list = c('AT.GC', 'CG.TA')
transition.column.list = c(1, 4)
transversion.list = c('AT.CG', 'AT.TA', 'CG.AT', 'CG.GC')
transversion.column.list = c(2, 3, 5, 6)


##GC content division
toGC.column.list = c(1, 2)
toAT.column.list = c(4, 5)

##all categories division
mutation.list  <- vector("list", 6)
mutation.list[[1]] = c('AT.GC')
mutation.list[[2]] = c('AT.CG')
mutation.list[[3]] = c('AT.TA') 
mutation.list[[4]] = c('CG.TA')
mutation.list[[5]] = c('CG.AT')
mutation.list[[6]] = c('CG.GC')

data<-read.table(input.name, sep="\t", header=T)

## makes into a list of data frames, divided up by line
structured.data = list();
line.list = levels(data$line)

structured.data <- vector("list", length(line.list))
for (i in 1:length(line.list)) {
	this.line <- line.list[i]
	print(this.line);
	structured.data[[i]] <- subset(data, line == this.line);
}

##get ancestral values for sites at risk from first genome
anc.genome.data <- structured.data[[1]][1,];
tot.mutations.signature <- c()
tot.mutations <- 0
for (i in 1:length(mutation.list)) {
  this.mutation.list<-mutation.list[i]
  tot.mutations.signature[i] <- 0  
	for (j in 1:length(this.mutation.list))
	{
    ## Total number of mutations at positions with data possible with this mutational signature in ANCESTOR
    column.name <- paste(this.mutation.list[j], ".anc.tot", sep="");
		tot.mutations.signature[i] <- tot.mutations.signature[i] + anc.genome.data[column.name][1,1];
	}  
  tot.mutations <- tot.mutations + tot.mutations.signature[i]
}
tot.base.pairs <- tot.mutations / 3;

## Calculated rates are per number of possible changes with signature
##
## To convert to a total per-genome rate,
## we need to know how many TOTAL bases there are of each type
## in the ANCESTRAL genome.

ml.best <- fit_ml_model(structured.data, mutation.list)


outfile <- file(description = output.name, open="w")


## Analyze bootstrap data sets to estimate 95% confidence limits.
## Calculated rates are per possible base change with signature per genome per generation
ml.distribution <- bootstrap_95_CI(structured.data, mutation.list, num.bootstrap.resamplings)


######## total rates ########

distribution.tot.rate = c();
for (j in 1:dim(ml.distribution)[1])
{
  distribution.tot.rate[j] <- 0;
  for (i in 1:length(mutation.list)) {
    distribution.tot.rate[j] <- distribution.tot.rate[j] + ml.distribution[j,i] * tot.mutations.signature[i]
  }
}
tot.rate.q <- quantile(distribution.tot.rate, pr=c(0.025, 0.975), type=1)

##############################
#### CL on compound rates ####
##############################

######## transitions ########

transition.tot.rate = c();
for (j in 1:dim(ml.distribution)[1])
{
  transition.tot.rate[j] <- 0;
  for (i in 1:length(transition.list)) {
    transition.tot.rate[j] <- transition.tot.rate[j] + ml.distribution[j,transition.column.list[i]]
  }
}
transition.q <- quantile(transition.tot.rate, pr=c(0.025, 0.975), type=1)

ml.transition.rate <- 0;
for (i in 1:length(transition.list)) {
  ml.transition.rate <- ml.transition.rate + ml.best[transition.column.list[i]] 
}

######## transversions ########

transversion.tot.rate = c();
for (j in 1:dim(ml.distribution)[1])
{
  transversion.tot.rate[j] <- 0;
  for (i in 1:length(transversion.list)) {
    transversion.tot.rate[j] <- transversion.tot.rate[j] + ml.distribution[j,transversion.column.list[i]]
  }
}
transversion.q <- quantile(transversion.tot.rate, pr=c(0.025, 0.975), type=1)

ml.transversion.rate <- 0;
for (i in 1:length(transversion.list)) {
  ml.transversion.rate <- ml.transversion.rate + ml.best[transversion.column.list[i]]
}

######## transition to transversion ratio ########
##Note: corrects for the fact that there are twice as many possible transversions as transitions
ttr <- transition.tot.rate / (transversion.tot.rate/2)
ttr.q <- quantile(ttr, pr=c(0.025, 0.975), type=1)
ml.ttr = ml.transition.rate / (ml.transversion.rate/2)


######## GC -> AT ratio ########
### Note that we are not accounting for base composition in this calculation (it assumes equal GC:AT)

toGC.tot.rate = c();
toAT.tot.rate = c();

for (j in 1:dim(ml.distribution)[1])
{
  toGC.tot.rate[j] <- 0;
  for (i in 1:length(toGC.column.list)) {
    toGC.tot.rate[j] <- toGC.tot.rate[j] + ml.distribution[j,toGC.column.list[i]]
  }
  
  toAT.tot.rate[j] <- 0;
  for (i in 1:length(toAT.column.list)) {
    toAT.tot.rate[j] <- toAT.tot.rate[j] + ml.distribution[j,toAT.column.list[i]]
  }
}

toAT.q <- quantile(toAT.tot.rate, pr=c(0.025, 0.975), type=1)
toGC.q <- quantile(toGC.tot.rate, pr=c(0.025, 0.975), type=1)


ml.toGC <- 0
for (i in 1:length(toGC.column.list)) {
	ml.toGC  <- ml.toGC + ml.best[toGC.column.list[i]]
}

ml.toAT <- 0
for (i in 1:length(toAT.column.list)) {
	ml.toAT  <- ml.toAT + ml.best[toAT.column.list[i]]
}  

ml.toAT.ratio = ml.toAT / ml.toGC

toAT.ratio = toAT.tot.rate / toGC.tot.rate;
toAT.ratio.q <- quantile(toAT.ratio, pr=c(0.025, 0.975), type=1)

#######################
######  Output   ######
#######################

## Print per bp change rates
cat("Rates \n");
cat("Type\tML-best\tLower-95%CI\tUpper-95%CI\n", file=outfile)
ml.tot.rate <- 0
for (i in 1:length(mutation.list)) {
	q <- quantile(ml.distribution[,i], pr=c(0.025, 0.975), type=1)
  	ml.tot.rate <- ml.tot.rate + ml.best[i] * tot.mutations.signature[i];
	cat( mutation.list[[i]], "\t", ml.best[i] , "\t", q[1], "\t", q[2], "\n", file=outfile );
}
## Print transition/transversion rates
cat( "transitions", "\t", ml.transition.rate , "\t", transition.q [1], "\t", transition.q [2], "\n", file=outfile );
cat( "transversions", "\t", ml.transversion.rate , "\t", transversion.q [1], "\t", transversion.q [2], "\n", file=outfile );

## Print transition/transversion rates
cat( "toGC", "\t", ml.toGC , "\t", toGC.q [1], "\t", toGC.q [2], "\n", file=outfile );
cat( "toAT", "\t", ml.toAT , "\t", toAT.q [1], "\t", toAT.q [2], "\n", file=outfile );

cat("\n", file=outfile)
cat("genome positions: ", tot.base.pairs, "\n", file=outfile);
cat("*** bootstrap ***", "\n", file=outfile)
cat("number of  replicates: ", num.bootstrap.resamplings, "\n", file=outfile)
cat("rate per genome [95% confidence]: ", ml.tot.rate, " [", tot.rate.q[1], ",", tot.rate.q[2], "]\n", file=outfile);
cat("rate per base pair [95% confidence]: ", ml.tot.rate/tot.base.pairs, " [", tot.rate.q[1]/tot.base.pairs, ",", tot.rate.q[2]/tot.base.pairs, "]\n", file=outfile);
cat("transition to transversion ratio [95% confidence]: ", ml.ttr, " [", ttr.q[1], ",", ttr.q[2], "]\n", file=outfile);
cat("toAT to toGC ratio [95% confidence]: ", ml.toAT.ratio, " [", toAT.ratio.q[1], ",", toAT.ratio.q[2], "]\n", file=outfile);

jackknife_95_CI(structured.data, mutation.list, tot.mutations.signature, outfile)
